import { Component, inject, Inject, Input, OnInit, ViewChild } from '@angular/core';
import { APP_LOCALES, LocaleConfig, Settings } from '@app/settings/settings.model';
import { AppPage, AppPageState } from '@app/shared/pages/base-page.class';
import { NetworkService } from '@app/network/network.service';
import { fadeInAnimation } from '@app/shared/animations';
import { Router } from '@angular/router';
import { RxStateProperty } from '@app/shared/decorator/state.decorator';
import { RxState } from '@rx-angular/state';
import { TranslateService } from '@ngx-translate/core';
import { IonContent } from '@ionic/angular';
import { isNotNilOrBlank } from '@app/shared/functions';

export interface HelpPageState extends AppPageState, Settings {}

export declare type HighlightedDefinition = 'GLOSSARY_DISTANCE_RULE' | 'DEMO_MODE_HELP' | 'READONLY_MODE_HELP' | 'UD' | 'login-method' | '';

@Component({
  selector: 'app-help',
  templateUrl: './help.page.html',
  styleUrls: ['./help.page.scss'],
  animations: [fadeInAnimation],
  providers: [RxState],
})
export class HelpPage extends AppPage<HelpPageState> implements OnInit {
  @ViewChild(IonContent, { static: true }) content: IonContent;
  @RxStateProperty() locale: string;

  @Input() highlightedDefinition: HighlightedDefinition;

  protected translate = inject(TranslateService);

  constructor(
    protected networkService: NetworkService,
    protected router: Router,
    @Inject(APP_LOCALES) public locales: LocaleConfig[]
  ) {
    super({ name: 'help' });
  }

  protected async ngOnLoad() {
    await Promise.all([this.networkService.ready()]);

    if (isNotNilOrBlank(this.highlightedDefinition)) {
      this.scrollToDefinition(this.highlightedDefinition);
    }
    return {
      ...this.settings.clone(),
    };
  }

  isDefinitionGlossaryDistanceRule() {
    return this.highlightedDefinition === 'GLOSSARY_DISTANCE_RULE';
  }

  isDemo() {
    return this.highlightedDefinition === 'DEMO_MODE_HELP';
  }

  isReadonly() {
    return this.highlightedDefinition === 'READONLY_MODE_HELP';
  }

  isUDHelp() {
    return this.highlightedDefinition === 'UD';
  }

  isLoginMethodHelp() {
    return this.highlightedDefinition === 'login-method';
  }

  async scrollToDefinition(definitionId: string) {
    const targetElement = document.getElementById(definitionId);
    if (targetElement) {
      const scrollElement = await this.content.getScrollElement();
      const yOffset = targetElement.getBoundingClientRect().top - scrollElement.getBoundingClientRect().top;
      this.content
        .scrollToPoint(0, yOffset, 500)
        .then(() => {
          console.log(`${this._logPrefix}Scrolled to ${definitionId}`);
        })
        .catch((err) => {
          console.error(`${this._logPrefix}Scroll error:`, err);
        });
    }
  }
}
